require 'morq/base'
require 'morq/list'
require 'morq/label'
require 'morq/maildir'
require 'morq/thread'
require 'morq/query'
require 'morq/spam'

require 'tmail'
require 'date'
require 'time'
require 'fileutils'
require "iconv"
require "nkf"
require "rast"

# INC:
# - allocate morq-id (32bit int)
# - link incoming mail to path (mid[3]/mid[0]/mid)
# - update index database (mid -> flags, eid, tid, date, from, subject)
# - update thread database (message-id -> mid / mid -> tid / tid -> mid...)
# - update inverted index
# - LABEL_ADD(mid, *ARGV)
# - LABEL_ADD(mid, inbox) unless ARGV include? "trash" nor "spam" nor "archive"
# - remove incoming file (link in path remains)

class Morq::Inc
  include Morq::Common
  include Morq::Thread
  include Morq::Label
  include Morq::Query
  include Morq::Spam

  private

  def allocate_mid
    path = morq_file("mid")
    FileUtils.mkpath(File.dirname(path))
    begin
      f = open(path, "r+")
      f.flock(File::LOCK_EX)
      data = f.read
      if data.nil? or data == ""
        num = 1                 # skip zero
      else
        num = unpack_id(data)
      end
      num += 1
    rescue Errno::ENOENT
      f = open(path, "w")
      num = 1
    end
    f.rewind
    f.write(pack_id(num))
    f.close
    num
  end

  def inc_mail(input)
    tmpdir = morq_file(""); FileUtils.mkpath(tmpdir)
    newdir = morq_file("spool"); FileUtils.mkpath(newdir)
    mdfile = Morq::Maildir::store(input, newdir, tmpdir)
    mdfile = File.join(newdir, mdfile)
    begin
      mid = allocate_mid()
      path = mid_path(mid)
      FileUtils.mkpath(File.dirname(path))
      File.link(mdfile, path)
    rescue
      retry
    end
    File.unlink(mdfile)
    mid
  end

  def classify_mail(mid, headers)
    labels = []
    path = mid_path(mid)
    if spam?(path)
      labels.push "spam"
    end
    labels
  end

  def read_headers(head)
    headers = Hash.new
    attr = nil
    head.each_line do |line|
      line.chop!
      next if /^From /=~line	# skip From-line
      break if /^$/=~line	# end of header
      line.gsub!(/\t/, " ")

      if /^(\S+?):\s*(.*)/ =~ line
        (attr = $1).downcase!
        if headers[attr]
          headers[attr] += "\n"
          headers[attr] += $2
        else
          headers[attr] = $2
        end
      elsif attr
        line.sub!(/^[\t\s]*/, '')
        headers[attr] += " " + line
      end
    end
    headers.default = ""
    headers
  end

  def decode_encoded_word(s)
    return NKF.nkf("-e", s)
  end

  def get_header_field(mail, field)
    mail[field]
  end

  def get_decoded_subject(mail)
    return decode_encoded_word(get_header_field(mail, 'subject'))
  end

  def extract_properties(mail, properties)
    for field in ["from", "to", "cc"]
      properties[field] = get_header_field(mail, field)
    end
    return properties
  end

  def extract_text(m)
    text = m["__text__"]
    mail = TMail::Mail.parse(text)
    if mail.multipart?
      return mail.parts.collect { |part|
        extract_text(part)
      }.join("\n")
    else
      case mail.content_type("text/plain")
      when "text/plain"
        return get_body(mail)
      when "text/html", "text/xml"
        return get_body(mail).gsub(/<.*?>/um, "")
      else
        return ""
      end
    end
  end

  def encode_internal(src, charset)
    begin
      return Iconv.conv("euc-jp", charset, src)
    rescue
      return NKF.nkf("-m0 -e", src)
    end
  end

  def get_body(mail)
    charset = mail.type_param("charset", @default_charset)
    return encode_internal(mail.body, charset)
  end

  def parse_mail(mid)
    text = File.read(mid_path(mid))
    header, = *text.split(/^\r\n/)
    m = read_headers(header)
    m["__text__"] = text
    m
  end

  def register_index(mid)
    m = parse_mail(mid)
    pid, tid = register_thread(mid, m)
    date = Time.parse(get_header_field(m, 'date')) rescue 0
    from = get_header_field(m, 'from')
    from = TMail::Address.parse(from).spec rescue from
    subject = get_decoded_subject(m)
    index = [0, pid, tid, date.to_i, from, subject]
    open_morq_db("index", true) do |db|
      db[pack_id(mid)] = pack_index(index)
    end
    m
  end

  def inverted_index_mail(mid, mail, extern=nil)
    path = mid_path(mid)
    index = nil
    open_morq_db("index") do |db|
      index = unpack_index(db[pack_id(mid)])
    end

    properties = {
      "filename" => path,
      "subject" => index[5],
      "from" => index[4],
      "date" => Time.at(index[3]).strftime("%FT%T%Z"),
      "mid" => mid,
    }
    text = extract_text(mail)
    properties = extract_properties(mail, properties)
    add_inverted_index(text, properties, extern)
  end

  public

  def inverted_index_mid(mid, db=nil)
    path = mid_path(mid)
    return unless File.exist?(path)
    inverted_index_mail(mid, parse_mail(mid), db)
  end

  def run(input, *labels)
    mid = inc_mail(input)
    mail = register_index(mid)
    if labels.size == 0
      labels = classify_mail(mid, mail)
    end
    open_label_db(:update) do |db|
      for label in labels
        db.add(mid, label)
      end
      unless labels.include? "spam" or labels.include? "trash" or labels.include? "archive"
        db.add(mid, "inbox")
      end
    end
    unless labels.include? "spam" or labels.include? "trash"
      inverted_index_mail(mid, mail)
    end
  end
end

if __FILE__ == $0
  inc = Morq::Inc.new
  inc.run(STDIN, *ARGV)
end
